import importlib.metadata

__version__ = importlib.metadata.version("aiowire")

from .event_loop import EventLoop, UnhandledException
from .poller import Poller
from .wire import (
    Wire,
    Sequence,
    Call,
    Repeat,
    Forever,
)



from typing import Optional, Callable, List, Dict
from inspect import isawaitable
import asyncio

from .wire import Wire

Handler = Callable[['EventLoop', Exception], None]

class UnhandledException(Exception):
    pass

def default_handler(handler : Optional[Handler] = None) -> Handler:
    """ This default handler captures the traceback that lead to its creation.

        Any exception raised by the provided handler will
        be re-raised, with an extra note on where this
        default_handler was created (i.e. the original `EventLoop.start()`)..

        If handler is None, then this works as if handler was just
        `raise UnhandledException`.
    """
    def insulated_handler(ev : 'EventLoop', e : Exception):
        if handler is None:
            raise UnhandledException("Wire with no exception handler") from e
        else:
            try:
                handler(ev, e)
            except Exception as e2:
                raise e2 from e
    return insulated_handler

class EventLoop:
    """
    Create a wire-driven event loop.

    This event loop manages a set of concurrent
    async tasks.

    New async tasks can be launched by calling
    `EventLoop.start`.

    The tasks must be async-functions which take the event-loop
    as its only parameter.

    The return values of these tasks must be either
    None, or else another `async function` which will
    be called with the current event loop as its argument.

    This way, an async function can be run repeatedly just
    by returning its own name.

    The intended use case is for a server that starts
    several network connections and subprocesses, and
    wants to perform task/connection management as
    part of some of the callbacks.

    Exception Handling:

    The ev.start() method can take a second parameter
    that is a callback where all exceptions are sent.
    When a task (or any of its subtasks) raise an exception,
    the callback is called with the exception as its argument.
    Note that if a task is cancelled, the
    exception type will be asyncio.CancelledError.
    Because each task either raises an exception,
    returns another wire, or returns None,
    the handler will only be called once by each start()-ed
    task (or not at all).
    """
    def __init__(self, timeout : Optional[float] = None):
        self.tasks : Dict[asyncio.Task, Handler] = {}
        self.timeout = timeout

    def start(self, w : Optional[Wire],
              handler : Optional[Handler] = None,
              capture_context : bool = True) -> None:
        """ Schedule the wire, `w`, for execution with
            the given exception handler.

            If capture_context is True, then exceptions raised
            during invocation of the handler will be wrapped with
            a traceback showing where start() was originally called.

            `capture_context` is set to False when running Wire-s returned
            by Wire-s so we don't trace the entire state history, only
            the original `start()` location.
        """
        if w is None:
            return None
        coro = w(self)
        if isawaitable(coro):
            #t = asyncio.ensure_future(coro)
            t : asyncio.Task = asyncio.create_task(coro) # type: ignore[arg-type]
            if handler is None or capture_context:
                self.tasks[t] = default_handler(handler)
            else:
                self.tasks[t] = handler
        return None

    async def run(self, timeout : Optional[float] = None) -> None:
        """
        Run the event loop.  Usually this is called
        automatically when the ``async with EventLoop ...``
        context ends.
        """
        loop = asyncio.get_running_loop()
        t0 = loop.time()
        t1 = t0
        if timeout is None:
            fin = None
        else:
            fin = t0+timeout
        while len(self.tasks) > 0 and (fin is None or t1 < fin):
            if fin is None:
                dt = None
            else:
                dt = fin - t1
            done, pending = await asyncio.wait(
                    self.tasks,
                    timeout = dt,
                    return_when = asyncio.FIRST_COMPLETED)
            for t in done:
                handler = self.tasks.pop(t)
                # Need to await t to get its return value,
                # then pass it to start again.
                try:
                    ret = await t
                    if isinstance(ret, Wire):
                        self.start(ret, handler, False)
                    elif ret is not None:
                        handler(self, TypeError(f"Wire returned {ret}"))
                except Exception as e:
                    handler(self, e)
            t1 = loop.time()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, traceback):
        # (exc_type, exc, traceback) are *sys.exc_info()
        # or else None (on normal exit)
        if exc_type is None: # run the event loop here
            await self.run(self.timeout)

        for t in self.tasks: #[max(self.cur-1,0):]:
            if not t.done():
                t.cancel()
        self.tasks.clear()
        return False # continue to raise any exception



from typing import Optional, Dict, Union

try:
    import zmq
    import zmq.asyncio 
    Socket = Union[zmq.Socket, int]
    zmqPOLLIN = zmq.POLLIN
except ImportError:
    zmq = None # type: ignore[assignment]
    Socket = int # type: ignore[misc]
    zmqPOLLIN = 101

from .wire import Wire
from .event_loop import EventLoop

class Poller(Wire):
    """
    File descriptor poller.  When a file it's watching
    gets input, it starts the corresponding callback (Wire).

    init takes a dictionary mapping sockets to Wire-s

    Interval is a report-back time (in milliseconds).
    It is only needed in the case that EventLoop is run without
    a timeout AND you have some other Wire that will call
    Poller.shutdown().

    See `the pyzmq docs <https://pyzmq.readthedocs.io/en/latest/api/zmq.html#polling>`_
    for more info.
    """
    def __init__(self, socks : Dict[Socket, Wire],
                       default_flags = zmqPOLLIN,
                       interval : Optional[int] = 1000):
        self.socks : Dict[Socket, Wire] = {}
        self.default_flags = default_flags
        self.interval = interval
        self.done = False

        self.poller = zmq.asyncio.Poller()
        for sock, cb in socks.items():
            self.register(sock, cb)

    def register(self, sock : Socket, cb : Wire, flags = None) -> None:
        """
        Add a listener on sock, invoking cb on activity.
        See pyzmq.Poller.register for more info on sock.

        If flags is None, self.default_flags is used.
        """
        if flags is None:
            flags = self.default_flags
        if sock in self.socks:
            raise IndexError(f"Already have a callback for sock: {sock}")
        self.poller.register(sock, flags)
        self.socks[sock] = cb

    def unregister(self, sock : Socket) -> None:
        self.poller.unregister(sock)
        del self.socks[sock]

    def shutdown(self) -> None:
        """
        Shutdown only needs to be called if the EventLoop
        is running in non-stop mode.
        """
        self.done = True
    
    async def __call__(self, ev : EventLoop) -> Optional[Wire]:
        # TODO: we could further capture the coroutine
        # here within self, then cancel it
        # during shutdown -- then await would probably throw something.
        events = await self.poller.poll(self.interval)
        if self.done:
            return None

        for fd, event in events:
            cb = self.socks.get(fd, None)
            if cb is not None:
                ev.start(cb)
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiowire']

package_data = \
{'': ['*']}

install_requires = \
['pyzmq>=23.2.0']

setup_kwargs = {
    'name': 'aiowire',
    'version': '3.0.1',
    'description': 'A simple event loop using asyncio',
    'long_description': 'aiowire - A simple event loop using asyncio\n============================================\n\n.. image:: https://github.com/frobnitzem/aiowire/actions/workflows/python-package.yml/badge.svg\n   :target: https://github.com/frobnitzem/aiowire/actions\n   :alt: CI\n.. image:: https://codecov.io/github/frobnitzem/aiowire/branch/main/graph/badge.svg\n   :target: https://app.codecov.io/gh/frobnitzem/aiowire\n   :alt: Coverage\n\nThis package implements a ``EventLoop`` class\nthat manages concurrent coroutines.\n\nIt is based on the principles of functional\nreactive programming and draws inspiration\nfrom Haskell\'s `Control.Wire <https://hackage.haskell.org/package/netwire-4.0.7/docs/Control-Wire.html>`_ library.\n\nIn particular, every co-routine started by the event loop is a ``Wire``.\n\n``Wire``-s either return ``None``, indicating they\'re done, or another\n``Wire``.\n\nAn example helps explain the idea::\n\n    from aiowire import EventLoop\n\n    event = 0\n    async def show_event(ev) -> Optional[Wire]:\n        print("Running...")\n        event += 1\n        await asyncio.sleep(event*0.15)\n        print(f"Event {event}")\n        if event < 5:\n            return Wire(show_event)\n        return None\n\n    async with EventLoop(timeout=1) as event:\n        event.start(show_event)\n        event.start(show_event)\n\n\nWe start up an event loop and drop in two wires.\nEach runs, then returns the ``show_event`` function.\nThe event loop runs those functions next... and so on.\n\nBut since this isn\'t functional programming.  The wires\nhave access to the event loop, and can start more\ntasks.  Easy, right?\n\n\nWhat can I do with it?\n^^^^^^^^^^^^^^^^^^^^^^\n\nWhat if you have a server that\'s spawning programs,\nworking with sockets, and managing timeouts?  Drop\nin one wire for each program, one polling on socket I/O,\nand another acting as a timer (as above).\n\nSome canonical task types that do these include::\n\n    asyncio.create_subprocess_exec # run a program\n\n    asyncio.sleep # awake the loop after a given time lapse\n\n    zmq.asyncio.Poller.poll # awake the loop after I/O on socket/file\n\n    aiowire.Poller # Wire-y interface to zmq.asyncio.Poller\n\n\nThink about each wire as a finite state machine.\nFor example,\n\n.. code::\n\n    flowchart LR\n        R[Ready] --> N{New Task?};\n        N -- Yes --> W[Working];\n        W --> C{Complete?};\n        C -- Yes --> R;\n\ncan be implemented like so::\n\n    async def ready(ev : EventLoop, info : X) -> Optional[Wire]:\n        if info.new_task():\n            do_working_action()\n            return Wire(working, info) # move to working state\n\n        # Return a sequence of 2 wires:\n        return Call(asyncio.sleep, 1.0) >> Wire(ready, info)\n\n    async def working(ev : EventLoop, info : X) -> Wire:\n        if info.complete():\n            do_complete_action()\n            return Wire(ready, info)\n        await asyncio.sleep(0.5) # directly sleep a bit\n        return Wire(working, info)\n\nNote how your sockets can launch programs, and your program\nresults can start/stop sockets, and everyone can start\nbackground tasks.\n\n\nPoller?\n^^^^^^^\n\nThe ``Poller`` class lets you schedule callbacks in response\nto socket or file-descriptor activity.  Of course, the callbacks\nare wires, and run concurrently.\n\nPoller is also a Wire, created as,\n`Poller(dictionary mapping sockets / fd-s to callback wires)`.\n\nYou add it to your event loop as usual::\n\n    # ... create sock from zmq.asyncio.Context\n\n    async def echo(ev):\n        await sock.send( await sock.recv() )\n\n    todo = { 0:  Call(print, "received input on sys.stdin"),\n             sock: Wire(echo)\n           }\n    async with EventLoop() as ev:\n        ev.start( Poller(todo) )\n\n\nTell me more\n^^^^^^^^^^^^\n\nYes, you *could* just send async functions taking one\nargument to ``EventLoop.start``, but where\'s the fun in\nwriting closures everywhere?\n\nTo take it to the next level, aiowire comes with a\n``Wire`` convenience class that lets you write ``Wire``-s expressively.\nThe following class extensions help you make Wire-s out of common \nprogramming idioms:\n\n* `Wire(w)`: acts like an identity over "async func(ev):" functions\n* `Repeat(w, n)`: repeat wire ``w`` n times in a row\n* `Forever(w)`: repeat forever -- like `Repeat(w) * infinity`\n* `Call(fn, *args, **kargs)`: call fn (normal or async),\n  ignore the return, and exit\n\nConsider, for example, printing 4 alarms separated by some time interval::\n\n    from aiowire import EventLoop, Call\n\n    prog = ( Call(asyncio.sleep, 0.1) >> Call(print, \'beep\\a\') ) * 4\n\n    async with EventLoop() as ev:\n        ev.start(prog)\n\nReferences\n==========\n\n* https://pyzmq.readthedocs.io/en/latest/api/zmq.html#poller\n* https://pythontic.com/modules/select/poll\n* https://blog.tomecek.net/post/non-blocking-stdin-in-python/\n',
    'author': 'David M. Rogers',
    'author_email': 'predictivestatmech@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/frobnitzem/aiowire',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
